/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese, Dominik Smith [committer] */


#ifndef UNPACK_H
#define UNPACK_H 1

#include "CbmDigiTimeslice.h"

#include "TofReadoutConfig.h"
#include "Trd2dReadoutConfig.h"
#include "TrdReadoutConfig.h"

#include <sstream>
#include <vector>

#include "BmonReadoutConfig.h"
#include "MuchReadoutConfig.h"
#include "StsReadoutConfigLegacy.h"
#include "UnpackBmon.h"
#include "UnpackMuch.h"
#include "UnpackSts.h"
#include "UnpackTof.h"
#include "UnpackTrd.h"
#include "UnpackTrd2d.h"

namespace cbm::algo
{

  /** @struct UnpackMonitorData
   ** @author Dominik Smith <d.smith@gsi.de>
   ** @since 2 Jun 2023
   ** @brief Monitoring data for unpacking
   **/
  struct UnpackMonitorData {
    std::vector<UnpackStsMonitorData> fSts;      ///< Monitoring data for STS
    std::vector<UnpackMuchMonitorData> fMuch;    ///< Monitoring data for MUCH
    std::vector<UnpackTofMonitorData> fTof;      ///< Monitoring data for TOF
    std::vector<UnpackBmonMonitorData> fBmon;    ///< Monitoring data for T0
    std::vector<UnpackTrdMonitorData> fTrd;      ///< Monitoring data for TRD
    std::vector<UnpackTrd2dMonitorData> fTrd2d;  ///< Monitoring data for TRD2D
    size_t fNumMs       = 0;
    size_t fNumBytes    = 0;
    size_t fNumDigis    = 0;
    size_t fNumCompUsed = 0;
    std::string print()
    {
      std::stringstream ss;
      ss << "TS stats: num MS " << fNumMs << ", bytes " << fNumBytes << ", digis " << fNumDigis << ", components "
         << fNumCompUsed << std::endl;
      return ss.str();
    }
  };

  /** @class Unpack
 ** @brief Algo class for unpacking digi timeslicess
 ** @author Dominik Smith <d.smith@gsi.de>
 ** @since 02.06.2023
 **
 **/
  class Unpack {

  public:
    typedef std::pair<CbmDigiTimeslice, UnpackMonitorData> resultType;

    /** @brief Algorithm execution
     ** @param fles timeslice to unpack
     ** @return pair: digi timeslice, monitoring data
     **/
    resultType operator()(const fles::Timeslice* timeslice);

    /** @brief Default constructor **/
    Unpack() {};

    /** @brief Destructor **/
    ~Unpack() {};

    /** @brief Parameters for STS unpackers **/
    StsReadoutConfigLegacy fStsConfig {};

    /** @brief Parameters for MUCH unpackers **/
    MuchReadoutConfig fMuchConfig {};

    /** @brief Parameters for TOF unpackers **/
    TofReadoutConfig fTofConfig {};

    /** @brief Parameters for T0 unpackers **/
    BmonReadoutConfig fBmonConfig {};

    /** @brief Parameters for TRD unpackers **/
    TrdReadoutConfig fTrdConfig {};

    /** @brief Parameters for TRD2D unpackers **/
    Trd2dReadoutConfig fTrd2dConfig {};

    /** @brief Initialize unpackers and fill parameters from config objects **/
    bool Init();

  private:  // methods
    /** @brief Microslice loop **/
    template<class Digi, class UnpackAlgo, class MonitorData>
    void MsLoop(const fles::Timeslice* timeslice, std::map<uint16_t, UnpackAlgo>& algoMap, const uint64_t comp,
                const uint16_t eqId, std::vector<Digi>* digis, UnpackMonitorData& monitor,
                std::vector<MonitorData>* monitorMs, uint8_t sys_ver);


  private:  // members
    /** @brief STS unpackers **/
    std::map<uint16_t, UnpackSts> fAlgoSts = {};

    /** @brief MUCH unpackers **/
    std::map<uint16_t, UnpackMuch> fAlgoMuch = {};

    /** @brief TOF unpackers **/
    std::map<uint16_t, UnpackTof> fAlgoTof = {};

    /** @brief T0 unpackers **/
    std::map<uint16_t, UnpackBmon> fAlgoBmon = {};

    /** @brief TRD unpackers **/
    std::map<uint16_t, UnpackTrd> fAlgoTrd = {};

    /** @brief TRD2D unpackers **/
    std::map<uint16_t, UnpackTrd2d> fAlgoTrd2d = {};
  };
}  // namespace cbm::algo

#endif /* UNPACK_H */
